@@ -24,7 +24,6 @@ from IPython.core.profiledir import ProfileDir
 from IPython.core.shellapp import (
     InteractiveShellApp, shell_flags, shell_aliases
 )
-from IPython.utils import io
 from ipython_genutils.path import filefind, ensure_dir_exists
 from traitlets import (
     Any, Instance, Dict, Unicode, Integer, Bool, DottedObjectName, Type, default
@@ -139,6 +138,7 @@ class IPKernelApp(BaseIPythonApplication, InteractiveShellApp,
     # streams, etc.
     no_stdout = Bool(False, help="redirect stdout to the null device").tag(config=True)
     no_stderr = Bool(False, help="redirect stderr to the null device").tag(config=True)
+    quiet = Bool(True, help="Only send stdout/stderr to output stream").tag(config=True)
     outstream_class = DottedObjectName('ipykernel.iostream.OutStream',
         help="The importstring for the OutStream factory").tag(config=True)
     displayhook_class = DottedObjectName('ipykernel.displayhook.ZMQDisplayHook',
@@ -165,7 +165,10 @@ class IPKernelApp(BaseIPythonApplication, InteractiveShellApp,
         if sys.platform == 'win32':
             if self.interrupt or self.parent_handle:
                 self.poller = ParentPollerWindows(self.interrupt, self.parent_handle)
-        elif self.parent_handle:
+        elif self.parent_handle and self.parent_handle != 1:
+            # PID 1 (init) is special and will never go away,
+            # only be reassigned.
+            # Parent polling doesn't work if ppid == 1 to start with.
             self.poller = ParentPollerUnix()
 
     def _bind_socket(self, s, port):
@@ -245,6 +248,14 @@ class IPKernelApp(BaseIPythonApplication, InteractiveShellApp,
         self.control_port = self._bind_socket(self.control_socket, self.control_port)
         self.log.debug("control ROUTER Channel on port: %i" % self.control_port)
 
+        if hasattr(zmq, 'ROUTER_HANDOVER'):
+            # set router-handover to workaround zeromq reconnect problems
+            # in certain rare circumstances
+            # see ipython/ipykernel#270 and zeromq/libzmq#2892
+            self.shell_socket.router_handover = \
+                self.control_socket.router_handover = \
+                self.stdin_socket.router_handover = 1
+
         self.init_iopub(context)
 
     def init_iopub(self, context):
@@ -290,9 +301,9 @@ class IPKernelApp(BaseIPythonApplication, InteractiveShellApp,
         # also raw print to the terminal if no parent_handle (`ipython kernel`)
         # unless log-level is CRITICAL (--quiet)
         if not self.parent_handle and self.log_level < logging.CRITICAL:
-            io.rprint(_ctrl_c_message)
+            print(_ctrl_c_message, file=sys.__stdout__)
             for line in lines:
-                io.rprint(line)
+                print(line, file=sys.__stdout__)
 
         self.ports = dict(shell=self.shell_port, iopub=self.iopub_port,
                                 stdin=self.stdin_port, hb=self.hb_port,
@@ -311,8 +322,20 @@ class IPKernelApp(BaseIPythonApplication, InteractiveShellApp,
         """Redirect input streams and set a display hook."""
         if self.outstream_class:
             outstream_factory = import_item(str(self.outstream_class))
-            sys.stdout = outstream_factory(self.session, self.iopub_thread, u'stdout')
-            sys.stderr = outstream_factory(self.session, self.iopub_thread, u'stderr')
+            if sys.stdout is not None:
+                sys.stdout.flush()
+
+            e_stdout = None if self.quiet else sys.__stdout__
+            e_stderr = None if self.quiet else sys.__stderr__
+
+            sys.stdout = outstream_factory(self.session, self.iopub_thread,
+                                           u'stdout',
+                                           echo=e_stdout)
+            if sys.stderr is not None:
+                sys.stderr.flush()
+            sys.stderr = outstream_factory(self.session, self.iopub_thread,
+                                           u'stderr',
+                                           echo=e_stderr)
         if self.displayhook_class:
             displayhook_factory = import_item(str(self.displayhook_class))
             self.displayhook = displayhook_factory(self.session, self.iopub_socket)
@@ -356,6 +379,7 @@ class IPKernelApp(BaseIPythonApplication, InteractiveShellApp,
         kernel_factory = self.kernel_class.instance
 
         kernel = kernel_factory(parent=self, session=self.session,
+                                control_stream=control_stream,
                                 shell_streams=[shell_stream, control_stream],
                                 iopub_thread=self.iopub_thread,
                                 iopub_socket=self.iopub_socket,
@@ -450,7 +474,13 @@ class IPKernelApp(BaseIPythonApplication, InteractiveShellApp,
         # file is definitely available at the time someone reads the log.
         self.log_connection_info()
         self.init_io()
-        self.init_signal()
+        try:
+            self.init_signal()
+        except:
+            # Catch exception when initializing signal fails, eg when running the 
+            # kernel on a separate thread 
+            if self.log_level < logging.CRITICAL:
+                self.log.error("Unable to initialize signal:", exc_info=True)
         self.init_kernel()
         # shell init steps
         self.init_path()
@@ -470,8 +500,9 @@ class IPKernelApp(BaseIPythonApplication, InteractiveShellApp,
         if self.poller is not None:
             self.poller.start()
         self.kernel.start()
+        self.io_loop = ioloop.IOLoop.current()
         try:
-            ioloop.IOLoop.instance().start()
+            self.io_loop.start()
         except KeyboardInterrupt:
             pass
 
